@@ -31,6 +31,7 @@
 #include <gdk/gdkkeysyms.h>
 #include <gdk/gdkx.h>
 
+#include "gp-handle.h"
 #include "panel-applets-manager.h"
 #include "panel-bindings.h"
 #include "panel.h"
@@ -60,8 +61,6 @@ struct _PanelAppletFrameActivating {
 
 /* PanelAppletFrame implementation */
 
-#define HANDLE_SIZE 10
-
 struct _PanelAppletFramePrivate {
 	PanelWidget     *panel;
 	AppletInfo      *applet_info;
@@ -72,10 +71,7 @@ struct _PanelAppletFramePrivate {
 
 	gchar           *iid;
 
-	GtkAllocation    child_allocation;
-	GdkRectangle     handle_rect;
-
-	guint            has_handle : 1;
+	GtkWidget       *handle;
 
 	gboolean         moving_focus_out;
 };
@@ -89,243 +85,246 @@ static guint panel_applet_frame_signals [LAST_SIGNAL];
 
 G_DEFINE_TYPE_WITH_PRIVATE (PanelAppletFrame, panel_applet_frame, GTK_TYPE_EVENT_BOX)
 
-static gboolean
-panel_applet_frame_draw (GtkWidget *widget,
-                         cairo_t   *cr)
-{
-        PanelAppletFrame *frame = PANEL_APPLET_FRAME (widget);
-	GtkStyleContext *context;
-	GtkStateFlags     state;
-	cairo_pattern_t  *bg_pattern;
-
-        if (GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->draw)
-                GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->draw (widget, cr);
-
-	if (!frame->priv->has_handle)
-		return FALSE;
-
-	context = gtk_widget_get_style_context (widget);
-	state = gtk_widget_get_state_flags (widget);
-	gtk_style_context_save (context);
-	gtk_style_context_set_state (context, state);
-
-	cairo_save (cr);
-
-	/* Set the pattern transform so as to correctly render a patterned
-	 * background with the handle */
-	gtk_style_context_get (context, state,
-			       "background-image", &bg_pattern,
-			       NULL);
-
-	if (bg_pattern) {
-		cairo_matrix_t ptm;
-
-		cairo_matrix_init_translate (&ptm,
-					     frame->priv->handle_rect.x,
-					     frame->priv->handle_rect.y);
-		cairo_matrix_scale (&ptm,
-				    frame->priv->handle_rect.width,
-				    frame->priv->handle_rect.height);
-		cairo_pattern_set_matrix (bg_pattern, &ptm);
-		cairo_pattern_destroy (bg_pattern);
-	}
-
-	gtk_render_handle (context, cr,
-			   frame->priv->handle_rect.x,
-			   frame->priv->handle_rect.y,
-			   frame->priv->handle_rect.width,
-			   frame->priv->handle_rect.height);
-
-	cairo_restore (cr);
-
-	gtk_style_context_restore (context);
-
-        return FALSE;
-}
-
-static void
-panel_applet_frame_get_preferred_width(GtkWidget *widget, gint *minimal_width, gint *natural_width)
-{
-	PanelAppletFrame *frame;
-	GtkBin           *bin;
-	GtkWidget        *child;
-	guint             border_width;
-
-	frame = PANEL_APPLET_FRAME (widget);
-	bin = GTK_BIN (widget);
-
-	if (!frame->priv->has_handle) {
-		GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->get_preferred_width (widget, minimal_width, natural_width);
-		return;
-	}
-
-	child = gtk_bin_get_child (bin);
-	if (child && gtk_widget_get_visible (child))
-		gtk_widget_get_preferred_width (child, minimal_width, natural_width);
-
-	border_width = gtk_container_get_border_width (GTK_CONTAINER (widget));
-	*minimal_width += border_width;
-	*natural_width += border_width;
-
-	switch (frame->priv->orientation) {
-	case PANEL_ORIENTATION_TOP:
-	case PANEL_ORIENTATION_BOTTOM:
-		*minimal_width += HANDLE_SIZE;
-		*natural_width += HANDLE_SIZE;
-		break;
-	case PANEL_ORIENTATION_LEFT:
-	case PANEL_ORIENTATION_RIGHT:
-		break;
-	default:
-		g_assert_not_reached ();
-		break;
-	}
-}
-
-static void
-panel_applet_frame_get_preferred_height(GtkWidget *widget, gint *minimal_height, gint *natural_height)
-{
-	PanelAppletFrame *frame;
-	GtkBin           *bin;
-	GtkWidget        *child;
-	guint             border_width;
-
-	frame = PANEL_APPLET_FRAME (widget);
-	bin = GTK_BIN (widget);
-
-	if (!frame->priv->has_handle) {
-		GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->get_preferred_height (widget, minimal_height, natural_height);
-		return;
-	}
-
-	child = gtk_bin_get_child (bin);
-	if (child && gtk_widget_get_visible (child))
-		gtk_widget_get_preferred_height (child, minimal_height, natural_height);
-
-	border_width = gtk_container_get_border_width (GTK_CONTAINER (widget));
-	*minimal_height += border_width;
-	*natural_height += border_width;
-
-	switch (frame->priv->orientation) {
-	case PANEL_ORIENTATION_LEFT:
-	case PANEL_ORIENTATION_RIGHT:
-		*minimal_height += HANDLE_SIZE;
-		*natural_height += HANDLE_SIZE;
-		break;
-	case PANEL_ORIENTATION_TOP:
-	case PANEL_ORIENTATION_BOTTOM:
-		break;
-	default:
-		g_assert_not_reached ();
-		break;
-	}
-}
-
 static void
 panel_applet_frame_size_allocate (GtkWidget     *widget,
-				  GtkAllocation *allocation)
+                                  GtkAllocation *allocation)
 {
-	PanelAppletFrame *frame;
-	GtkBin           *bin;
-	GtkWidget        *child;
-	GdkWindow        *window;
-	GtkAllocation     new_allocation;
-	GtkAllocation     widget_allocation;
+  PanelAppletFrame *self;
+  GtkRequisition handle_size;
+  GtkAllocation handle_allocation;
+  GtkAllocation child_allocation;
+  GtkWidget *child;
 
-	frame = PANEL_APPLET_FRAME (widget);
-	bin = GTK_BIN (widget);
+  self = PANEL_APPLET_FRAME (widget);
 
-	gtk_widget_get_allocation (widget, &widget_allocation);
+  GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->size_allocate (widget,
+                                                                     allocation);
 
-	if (!frame->priv->has_handle) {
-		GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->size_allocate (widget,
-										   allocation);
-		return;
-	}
+  if (self->priv->handle == NULL ||
+      !gtk_widget_get_visible (self->priv->handle))
+    return;
 
-	gtk_widget_set_allocation (widget, allocation);
+  gtk_widget_get_preferred_size (self->priv->handle, &handle_size, NULL);
 
-	frame->priv->handle_rect.x = 0;
-	frame->priv->handle_rect.y = 0;
+  if (!gtk_widget_get_has_window (widget))
+    {
+      handle_allocation.x = allocation->x;
+      handle_allocation.y = allocation->y;
 
-	switch (frame->priv->orientation) {
-	case PANEL_ORIENTATION_TOP:
-	case PANEL_ORIENTATION_BOTTOM:
-		frame->priv->handle_rect.width  = HANDLE_SIZE;
-		frame->priv->handle_rect.height = allocation->height;
+      child_allocation.x = allocation->x;
+      child_allocation.y = allocation->y;
+    }
+  else
+    {
+      handle_allocation.x = 0;
+      handle_allocation.y = 0;
 
-		if (gtk_widget_get_direction (GTK_WIDGET (frame)) !=
-		    GTK_TEXT_DIR_RTL) {
-			frame->priv->handle_rect.x = 0;
-			new_allocation.x = HANDLE_SIZE;
-		} else {
-			frame->priv->handle_rect.x = allocation->width - HANDLE_SIZE;
-			new_allocation.x = 0;
-		}
+      child_allocation.x = 0;
+      child_allocation.y = 0;
+    }
 
-		new_allocation.y      = 0;
-		new_allocation.width  = allocation->width - HANDLE_SIZE;
-		new_allocation.height = allocation->height;
-		break;
-	case PANEL_ORIENTATION_LEFT:
-	case PANEL_ORIENTATION_RIGHT:
-		frame->priv->handle_rect.width  = allocation->width;
-		frame->priv->handle_rect.height = HANDLE_SIZE;
+  switch (self->priv->orientation)
+    {
+      case PANEL_ORIENTATION_TOP:
+      case PANEL_ORIENTATION_BOTTOM:
+        if (gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL)
+          handle_allocation.x = allocation->width - handle_size.width;
+        else
+          child_allocation.x += handle_size.width;
 
-		new_allocation.x      = 0;
-		new_allocation.y      = HANDLE_SIZE;
-		new_allocation.width  = allocation->width;
-		new_allocation.height = allocation->height - HANDLE_SIZE;
-		break;
-	default:
-		g_assert_not_reached ();
-		break;
-	}
+        handle_allocation.width = handle_size.width;
+        handle_allocation.height = allocation->height;
 
-	new_allocation.width  = MAX (1, new_allocation.width);
-	new_allocation.height = MAX (1, new_allocation.height);
+        child_allocation.width = allocation->width - handle_size.width;
+        child_allocation.height = allocation->height;
+        break;
 
-	window = gtk_widget_get_window (widget);
+      case PANEL_ORIENTATION_LEFT:
+      case PANEL_ORIENTATION_RIGHT:
+        handle_allocation.width = allocation->width;
+        handle_allocation.height = handle_size.height;
 
-	/* If the child allocation changed, that means that the frame is drawn
-	 * in a new place, so we must redraw the entire widget.
-	 */
-	if (gtk_widget_get_mapped (widget) &&
-	    (new_allocation.x != frame->priv->child_allocation.x ||
-	     new_allocation.y != frame->priv->child_allocation.y ||
-	     new_allocation.width != frame->priv->child_allocation.width ||
-	     new_allocation.height != frame->priv->child_allocation.height))
-		gdk_window_invalidate_rect (window, &widget_allocation, FALSE);
+        child_allocation.y += handle_size.height;
+        child_allocation.width = allocation->width;
+        child_allocation.height = allocation->height - handle_size.height;
+        break;
 
-	if (gtk_widget_get_realized (widget)) {
-		guint border_width;
+      default:
+        g_assert_not_reached ();
+        break;
+    }
 
-		border_width = gtk_container_get_border_width (GTK_CONTAINER (widget));
-		gdk_window_move_resize (window,
-			allocation->x + border_width,
-			allocation->y + border_width,
-			MAX (allocation->width - border_width * 2, 0),
-			MAX (allocation->height - border_width * 2, 0));
-	}
+  gtk_widget_size_allocate (self->priv->handle, &handle_allocation);
 
-	child = gtk_bin_get_child (bin);
-	if (child && gtk_widget_get_visible (child))
-		gtk_widget_size_allocate (child, &new_allocation);
-  
-	frame->priv->child_allocation = new_allocation;
+  child = gtk_bin_get_child (GTK_BIN (widget));
+  if (child != NULL && gtk_widget_get_visible (child))
+    gtk_widget_size_allocate (child, &child_allocation);
+}
+
+static void
+panel_applet_frame_get_preferred_height (GtkWidget *widget,
+                                         gint      *minimum_height,
+                                         gint      *natural_height)
+{
+  PanelAppletFrame *self;
+  int handle_minimum_height;
+  int handle_natural_height;
+
+  self = PANEL_APPLET_FRAME (widget);
+
+  GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->get_preferred_height (widget,
+                                                                            minimum_height,
+                                                                            natural_height);
+
+  if (self->priv->handle == NULL ||
+      !gtk_widget_get_visible (self->priv->handle))
+    return;
+
+  gtk_widget_get_preferred_height (self->priv->handle,
+                                   &handle_minimum_height,
+                                   &handle_natural_height);
+
+  if (self->priv->orientation & PANEL_VERTICAL_MASK)
+    {
+      *minimum_height += handle_minimum_height;
+      *natural_height += handle_natural_height;
+    }
+  else
+    {
+      *minimum_height = MAX (*minimum_height, handle_minimum_height);
+      *natural_height = MAX (*natural_height, handle_natural_height);
+    }
+}
+
+static void
+panel_applet_frame_get_preferred_width_for_height (GtkWidget *widget,
+                                                   gint       height,
+                                                   gint      *minimum_width,
+                                                   gint      *natural_width)
+{
+  PanelAppletFrame *self;
+  int handle_minimum_width;
+  int handle_natural_width;
+
+  self = PANEL_APPLET_FRAME (widget);
+
+  GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->get_preferred_width_for_height (widget,
+                                                                                      height,
+                                                                                      minimum_width,
+                                                                                      natural_width);
+
+  if (self->priv->handle == NULL ||
+      !gtk_widget_get_visible (self->priv->handle))
+    return;
+
+  gtk_widget_get_preferred_width_for_height (self->priv->handle,
+                                             height,
+                                             &handle_minimum_width,
+                                             &handle_natural_width);
+
+  if (self->priv->orientation & PANEL_HORIZONTAL_MASK)
+    {
+      *minimum_width += handle_minimum_width;
+      *natural_width += handle_natural_width;
+    }
+  else
+    {
+      *minimum_width = MAX (*minimum_width, handle_minimum_width);
+      *natural_width = MAX (*natural_width, handle_natural_width);
+    }
+}
+
+static void
+panel_applet_frame_get_preferred_width (GtkWidget *widget,
+                                        gint      *minimum_width,
+                                        gint      *natural_width)
+{
+  PanelAppletFrame *self;
+  int handle_minimum_width;
+  int handle_natural_width;
+
+  self = PANEL_APPLET_FRAME (widget);
+
+  GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->get_preferred_width (widget,
+                                                                           minimum_width,
+                                                                           natural_width);
+
+  if (self->priv->handle == NULL ||
+      !gtk_widget_get_visible (self->priv->handle))
+    return;
+
+  gtk_widget_get_preferred_width (self->priv->handle,
+                                  &handle_minimum_width,
+                                  &handle_natural_width);
+
+  if (self->priv->orientation & PANEL_HORIZONTAL_MASK)
+    {
+      *minimum_width += handle_minimum_width;
+      *natural_width += handle_natural_width;
+    }
+  else
+    {
+      *minimum_width = MAX (*minimum_width, handle_minimum_width);
+      *natural_width = MAX (*natural_width, handle_natural_width);
+    }
+}
+
+static void
+panel_applet_frame_get_preferred_height_for_width (GtkWidget *widget,
+                                                   gint       width,
+                                                   gint      *minimum_height,
+                                                   gint      *natural_height)
+{
+  PanelAppletFrame *self;
+  int handle_minimum_height;
+  int handle_natural_height;
+
+  self = PANEL_APPLET_FRAME (widget);
+
+  GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->get_preferred_height_for_width (widget,
+                                                                                      width,
+                                                                                      minimum_height,
+                                                                                      natural_height);
+
+  if (self->priv->handle == NULL ||
+      !gtk_widget_get_visible (self->priv->handle))
+    return;
+
+  gtk_widget_get_preferred_height_for_width (self->priv->handle,
+                                             width,
+                                             &handle_minimum_height,
+                                             &handle_natural_height);
+
+  if (self->priv->orientation & PANEL_VERTICAL_MASK)
+    {
+      *minimum_height += handle_minimum_height;
+      *natural_height += handle_natural_height;
+    }
+  else
+    {
+      *minimum_height = MAX (*minimum_height, handle_minimum_height);
+      *natural_height = MAX (*natural_height, handle_natural_height);
+    }
 }
 
 static inline gboolean
-button_event_in_rect (GdkEventButton *event,
-		      GdkRectangle   *rect)
+button_event_in_handle (PanelAppletFrame *self,
+                        GdkEventButton   *event)
 {
-	if (event->x >= rect->x &&
-	    event->x <= (rect->x + rect->width) &&
-	    event->y >= rect->y &&
-	    event->y <= (rect->y + rect->height))
-		return TRUE;
+  GtkAllocation allocation;
 
-	return FALSE;
+  if (self->priv->handle == NULL)
+    return FALSE;
+
+  gtk_widget_get_allocation (self->priv->handle, &allocation);
+
+  if (event->x >= allocation.x &&
+      event->x <= (allocation.x + allocation.width) &&
+      event->y >= allocation.y &&
+      event->y <= (allocation.y + allocation.height))
+    return TRUE;
+
+  return FALSE;
 }
 
 static void
@@ -439,7 +438,7 @@ panel_applet_frame_button_changed (GtkWidget      *widget,
 	switch (event->button) {
 	case 1:
 	case 2:
-		if (button_event_in_rect (event, &frame->priv->handle_rect)) {
+		if (button_event_in_handle (frame, event)) {
 			if ((event->type == GDK_BUTTON_PRESS ||
 			     event->type == GDK_2BUTTON_PRESS) &&
 			    modifiers == panel_bindings_get_mouse_button_modifier_keymask ()){
@@ -522,6 +521,48 @@ panel_applet_frame_focus (GtkWidget        *widget,
 	return GTK_WIDGET_CLASS (panel_applet_frame_parent_class)->focus (widget, dir);
 }
 
+static void
+panel_applet_frame_forall (GtkContainer *container,
+                           gboolean      include_internals,
+                           GtkCallback   callback,
+                           gpointer      callback_data)
+{
+  PanelAppletFrame *self;
+
+  self = PANEL_APPLET_FRAME (container);
+
+  GTK_CONTAINER_CLASS (panel_applet_frame_parent_class)->forall (container,
+                                                                 include_internals,
+                                                                 callback,
+                                                                 callback_data);
+
+  if (include_internals && self->priv->handle)
+    (* callback) (self->priv->handle, callback_data);
+}
+
+static void
+panel_applet_frame_remove (GtkContainer *container,
+                           GtkWidget    *widget)
+{
+  PanelAppletFrame *self;
+
+  self = PANEL_APPLET_FRAME (container);
+
+  if (self->priv->handle != widget)
+    {
+      GTK_CONTAINER_CLASS (panel_applet_frame_parent_class)->remove (container,
+                                                                     widget);
+
+      return;
+    }
+
+  gtk_widget_unparent (widget);
+  self->priv->handle = NULL;
+
+  if (gtk_widget_get_visible (GTK_WIDGET (container)))
+    gtk_widget_queue_resize (GTK_WIDGET (container));
+}
+
 static void
 add_tab_bindings (GtkBindingSet    *binding_set,
                   GdkModifierType   modifiers,
@@ -540,18 +581,25 @@ panel_applet_frame_class_init (PanelAppletFrameClass *klass)
 {
 	GObjectClass   *gobject_class = (GObjectClass *) klass;
 	GtkWidgetClass *widget_class = (GtkWidgetClass *) klass;
+	GtkContainerClass *container_class;
 	GtkBindingSet *binding_set;
 
+	container_class = GTK_CONTAINER_CLASS (klass);
+
 	gobject_class->finalize = panel_applet_frame_finalize;
 
-	widget_class->draw                 = panel_applet_frame_draw;
-	widget_class->get_preferred_width  = panel_applet_frame_get_preferred_width;
+	widget_class->size_allocate = panel_applet_frame_size_allocate;
 	widget_class->get_preferred_height = panel_applet_frame_get_preferred_height;
-	widget_class->size_allocate        = panel_applet_frame_size_allocate;
+	widget_class->get_preferred_width_for_height = panel_applet_frame_get_preferred_width_for_height;
+	widget_class->get_preferred_width = panel_applet_frame_get_preferred_width;
+	widget_class->get_preferred_height_for_width = panel_applet_frame_get_preferred_height_for_width;
 	widget_class->button_press_event   = panel_applet_frame_button_changed;
 	widget_class->button_release_event = panel_applet_frame_button_changed;
 	widget_class->focus                = panel_applet_frame_focus;
 
+	container_class->forall = panel_applet_frame_forall;
+	container_class->remove = panel_applet_frame_remove;
+
 	klass->move_focus_out_of_applet = panel_applet_frame_move_focus_out_of_applet;
 
 	panel_applet_frame_signals [MOVE_FOCUS_OUT_OF_APPLET] =
@@ -571,17 +619,11 @@ panel_applet_frame_class_init (PanelAppletFrameClass *klass)
 static void
 panel_applet_frame_init (PanelAppletFrame *frame)
 {
-	GtkStyleContext *context;
-
 	frame->priv = panel_applet_frame_get_instance_private (frame);
 
 	frame->priv->panel       = NULL;
 	frame->priv->orientation = PANEL_ORIENTATION_TOP;
 	frame->priv->applet_info = NULL;
-	frame->priv->has_handle  = FALSE;
-
-	context = gtk_widget_get_style_context (GTK_WIDGET (frame));
-	gtk_style_context_add_class (context, GTK_STYLE_CLASS_HORIZONTAL);
 }
 
 static void
@@ -596,7 +638,16 @@ update_flags (PanelAppletFrame *self)
   major = (flags & GP_APPLET_FLAGS_EXPAND_MAJOR) != 0;
   minor = (flags & GP_APPLET_FLAGS_EXPAND_MINOR) != 0;
 
-  self->priv->has_handle = (flags & GP_APPLET_FLAGS_HAS_HANDLE) != 0;
+  if ((flags & GP_APPLET_FLAGS_HAS_HANDLE) != 0)
+    {
+      self->priv->handle = gp_handle_new ();
+      gtk_widget_set_parent (self->priv->handle, GTK_WIDGET (self));
+      gtk_widget_show (self->priv->handle);
+    }
+  else
+    {
+      g_clear_pointer (&self->priv->handle, gtk_widget_destroy);
+    }
 
   panel_widget_set_applet_expandable (self->priv->panel,
                                       GTK_WIDGET (self),
@@ -612,12 +663,18 @@ update_size_hints (PanelAppletFrame *self)
 
   size_hints = gp_applet_get_size_hints (self->priv->applet, &n_elements);
 
-  if (self->priv->has_handle)
+  if (self->priv->handle != NULL)
     {
+      GtkAllocation allocation;
       gint extra_size;
       guint i;
 
-      extra_size = HANDLE_SIZE + 1;
+      gtk_widget_get_allocation (self->priv->handle, &allocation);
+
+      if (self->priv->orientation & PANEL_HORIZONTAL_MASK)
+        extra_size = allocation.width + 1;
+      else
+        extra_size = allocation.height + 1;
 
       for (i = 0; i < n_elements; i++)
         size_hints[i] += extra_size;
@@ -690,23 +747,11 @@ void
 panel_applet_frame_change_orientation (PanelAppletFrame *frame,
 				       PanelOrientation  orientation)
 {
-	GtkStyleContext *context;
-
 	if (orientation == frame->priv->orientation)
 		return;
 
 	frame->priv->orientation = orientation;
 
-	context = gtk_widget_get_style_context (GTK_WIDGET (frame));
-	if (orientation & PANEL_HORIZONTAL_MASK) {
-		gtk_style_context_add_class (context, GTK_STYLE_CLASS_HORIZONTAL);
-		gtk_style_context_remove_class (context, GTK_STYLE_CLASS_VERTICAL);
-	} else {
-		gtk_style_context_add_class (context, GTK_STYLE_CLASS_VERTICAL);
-		gtk_style_context_remove_class (context, GTK_STYLE_CLASS_HORIZONTAL);
-	}
-	gtk_widget_reset_style (GTK_WIDGET (frame));
-
 	frame_change_orientation (frame, orientation);
 }
 
@@ -749,6 +794,9 @@ _panel_applet_frame_set_applet (PanelAppletFrame *self,
                     "size-hints-changed",
                     G_CALLBACK (size_hints_changed_cb),
                     self);
+
+  gtk_container_add (GTK_CONTAINER (self), GTK_WIDGET (applet));
+  gtk_widget_show (GTK_WIDGET (applet));
 }
 
 void
