@@ -51,12 +51,6 @@ static GSList *panels = NULL;
 static gboolean panel_applet_in_drag = FALSE;
 static GtkWidget *saved_focus_widget = NULL;
 
-static void panel_widget_get_preferred_width  (GtkWidget        *widget,
-					       gint             *minimal_width,
-					       gint             *natural_width);
-static void panel_widget_get_preferred_height (GtkWidget        *widget,
-					       gint             *minimal_height,
-					       gint             *natural_height);
 static void panel_widget_size_allocate        (GtkWidget        *widget,
 					       GtkAllocation    *allocation);
 static void panel_widget_cadd                 (GtkContainer     *container,
@@ -64,7 +58,6 @@ static void panel_widget_cadd                 (GtkContainer     *container,
 static void panel_widget_cremove              (GtkContainer     *container,
 					       GtkWidget        *widget);
 static void panel_widget_dispose              (GObject          *obj);
-static void panel_widget_finalize             (GObject          *obj);
 
 static void panel_widget_push_move_applet   (PanelWidget      *panel,
                                              GtkDirectionType  dir);
@@ -258,6 +251,297 @@ remove_all_move_bindings (PanelWidget *panel)
 	gtk_binding_entry_remove (binding_set, GDK_KEY_space, 0);
 }
 
+static void
+get_preferred_size_from_hints (AppletData *ad,
+                               int        *minimum_size,
+                               int        *natural_size)
+{
+  int max_size;
+  int min_size;
+  int i;
+
+  max_size = 0;
+  min_size = INT_MAX;
+
+  for (i = 0; i < ad->size_hints_len; i += 2)
+    {
+      max_size = MAX (max_size, ad->size_hints[i]);
+      min_size = MIN (min_size, ad->size_hints[i + 1]);
+    }
+
+  *minimum_size = min_size;
+  *natural_size = max_size;
+}
+
+static GtkSizeRequestMode
+panel_widget_get_request_mode (GtkWidget *widget)
+{
+  PanelWidget *self;
+
+  self = PANEL_WIDGET (widget);
+
+  if (self->orient == GTK_ORIENTATION_VERTICAL)
+    return GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH;
+
+  return GTK_SIZE_REQUEST_WIDTH_FOR_HEIGHT;
+}
+
+static void
+panel_widget_get_preferred_height (GtkWidget *widget,
+                                   gint      *minimum_height,
+                                   gint      *natural_height)
+{
+  PanelWidget *self;
+  int min_height;
+  int nat_height;
+
+  self = PANEL_WIDGET (widget);
+
+  min_height = 0;
+  nat_height = 0;
+
+  if (self->orient == GTK_ORIENTATION_HORIZONTAL)
+    {
+      min_height = MAX (self->sz, 12);
+      nat_height = MAX (self->sz, 12);
+    }
+  else if (self->orient == GTK_ORIENTATION_VERTICAL)
+    {
+      GList *l;
+
+      for (l = self->applet_list; l != NULL; l = g_list_next (l))
+        {
+          AppletData *ad;
+          int child_min_height;
+          int child_nat_height;
+
+          ad = l->data;
+
+          gtk_widget_get_preferred_height (ad->applet,
+                                           &child_min_height,
+                                           &child_nat_height);
+
+          if (ad->expand_major)
+            {
+              if (ad->size_hints != NULL)
+                {
+                  get_preferred_size_from_hints (ad,
+                                                 &child_min_height,
+                                                 &child_nat_height);
+                }
+
+              min_height += child_min_height;
+              nat_height += child_nat_height;
+            }
+          else
+            {
+              min_height += child_min_height;
+              nat_height += child_min_height;
+            }
+        }
+    }
+  else
+    {
+      g_assert_not_reached ();
+    }
+
+  *minimum_height = min_height;
+  *natural_height = nat_height;
+}
+
+static void
+panel_widget_get_preferred_width_for_height (GtkWidget *widget,
+                                             gint       height,
+                                             gint      *minimum_width,
+                                             gint      *natural_width)
+{
+  PanelWidget *self;
+  int min_width;
+  int nat_width;
+
+  self = PANEL_WIDGET (widget);
+
+  min_width = 0;
+  nat_width = 0;
+
+  if (self->orient == GTK_ORIENTATION_VERTICAL)
+    {
+      min_width = MAX (self->sz, 12);
+      nat_width = MAX (self->sz, 12);
+    }
+  else if (self->orient == GTK_ORIENTATION_HORIZONTAL)
+    {
+      GList *l;
+
+      for (l = self->applet_list; l != NULL; l = g_list_next (l))
+        {
+          AppletData *ad;
+          int child_min_width;
+          int child_nat_width;
+
+          ad = l->data;
+
+          gtk_widget_get_preferred_width_for_height (ad->applet,
+                                                     height,
+                                                     &child_min_width,
+                                                     &child_nat_width);
+
+          if (ad->expand_major)
+            {
+              if (ad->size_hints != NULL)
+                {
+                  get_preferred_size_from_hints (ad,
+                                                 &child_min_width,
+                                                 &child_nat_width);
+                }
+
+              min_width += child_min_width;
+              nat_width += child_nat_width;
+            }
+          else
+            {
+              min_width += child_min_width;
+              nat_width += child_min_width;
+            }
+        }
+    }
+  else
+    {
+      g_assert_not_reached ();
+    }
+
+  *minimum_width = min_width;
+  *natural_width = nat_width;
+}
+
+static void
+panel_widget_get_preferred_width (GtkWidget *widget,
+                                  gint      *minimum_width,
+                                  gint      *natural_width)
+{
+  PanelWidget *self;
+  int min_width;
+  int nat_width;
+
+  self = PANEL_WIDGET (widget);
+
+  min_width = 0;
+  nat_width = 0;
+
+  if (self->orient == GTK_ORIENTATION_VERTICAL)
+    {
+      min_width = MAX (self->sz, 12);
+      nat_width = MAX (self->sz, 12);
+    }
+  else if (self->orient == GTK_ORIENTATION_HORIZONTAL)
+    {
+      GList *l;
+
+      for (l = self->applet_list; l != NULL; l = g_list_next (l))
+        {
+          AppletData *ad;
+          int child_min_width;
+          int child_nat_width;
+
+          ad = l->data;
+
+          gtk_widget_get_preferred_width (ad->applet,
+                                          &child_min_width,
+                                          &child_nat_width);
+
+          if (ad->expand_major)
+            {
+              if (ad->size_hints != NULL)
+                {
+                  get_preferred_size_from_hints (ad,
+                                                 &child_min_width,
+                                                 &child_nat_width);
+                }
+
+              min_width += child_min_width;
+              nat_width += child_nat_width;
+            }
+          else
+            {
+              min_width += child_min_width;
+              nat_width += child_min_width;
+            }
+        }
+    }
+  else
+    {
+      g_assert_not_reached ();
+    }
+
+  *minimum_width = min_width;
+  *natural_width = nat_width;
+}
+
+static void
+panel_widget_get_preferred_height_for_width (GtkWidget *widget,
+                                             gint       width,
+                                             gint      *minimum_height,
+                                             gint      *natural_height)
+{
+  PanelWidget *self;
+  int min_height;
+  int nat_height;
+
+  self = PANEL_WIDGET (widget);
+
+  min_height = 0;
+  nat_height = 0;
+
+  if (self->orient == GTK_ORIENTATION_HORIZONTAL)
+    {
+      min_height = MAX (self->sz, 12);
+      nat_height = MAX (self->sz, 12);
+    }
+  else if (self->orient == GTK_ORIENTATION_VERTICAL)
+    {
+      GList *l;
+
+      for (l = self->applet_list; l != NULL; l = g_list_next (l))
+        {
+          AppletData *ad;
+          int child_min_height;
+          int child_nat_height;
+
+          ad = l->data;
+
+          gtk_widget_get_preferred_height_for_width (ad->applet,
+                                                     width,
+                                                     &child_min_height,
+                                                     &child_nat_height);
+
+          if (ad->expand_major)
+            {
+              if (ad->size_hints != NULL)
+                {
+                  get_preferred_size_from_hints (ad,
+                                                 &child_min_height,
+                                                 &child_nat_height);
+                }
+
+              min_height += child_min_height;
+              nat_height += child_nat_height;
+            }
+          else
+            {
+              min_height += child_min_height;
+              nat_height += child_min_height;
+            }
+        }
+    }
+  else
+    {
+      g_assert_not_reached ();
+    }
+
+  *minimum_height = min_height;
+  *natural_height = nat_height;
+}
+
 static void
 panel_widget_class_init (PanelWidgetClass *class)
 {
@@ -357,10 +641,12 @@ panel_widget_class_init (PanelWidgetClass *class)
 	class->end_move = panel_widget_end_move;
 
 	object_class->dispose = panel_widget_dispose;
-	object_class->finalize = panel_widget_finalize;
-	
-	widget_class->get_preferred_width = panel_widget_get_preferred_width;
+
+	widget_class->get_request_mode = panel_widget_get_request_mode;
 	widget_class->get_preferred_height = panel_widget_get_preferred_height;
+	widget_class->get_preferred_width_for_height = panel_widget_get_preferred_width_for_height;
+	widget_class->get_preferred_width = panel_widget_get_preferred_width;
+	widget_class->get_preferred_height_for_width = panel_widget_get_preferred_height_for_width;
 	widget_class->size_allocate = panel_widget_size_allocate;
 	widget_class->focus = panel_widget_real_focus;
 
@@ -493,27 +779,6 @@ get_applet_list_pack (PanelWidget         *panel,
 	return ret;
 }
 
-static int
-get_size_from_hints (AppletData *ad, int cells)
-{
-	int i;
-	
-	for (i = 0; i < ad->size_hints_len; i += 2) {
-		if (cells > ad->size_hints[i]) {
-			/* Clip to top */
-			cells = ad->size_hints[i];
-			break;
-		}
-		if (cells <= ad->size_hints[i] &&
-		    cells >= ad->size_hints[i+1]) {
-			/* Keep cell size */
-			break;
-		}
-	}
-	
-	return MAX (cells, ad->min_cells);
-}
-
 static void
 panel_widget_compress_pack_indexes_list (PanelWidget *panel,
 					 GList       *list)
@@ -551,57 +816,8 @@ panel_widget_compress_pack_indexes (PanelWidget *panel)
 	g_list_free (list);
 }
 
-static void
-panel_widget_update_size_hints_for_toplevel (PanelWidget *panel)
-{
-	GList *list;
-	GList *ad_with_hints = NULL;
-
-	panel->nb_applets_size_hints = 0;
-	if (panel->applets_hints != NULL)
-		g_free (panel->applets_hints);
-	panel->applets_hints = NULL;
-	if (panel->applets_using_hint != NULL)
-		g_free (panel->applets_using_hint);
-	panel->applets_using_hint = NULL;
-
-	if (!panel->packed)
-		return;
-
-	for (list = panel->applet_list; list; list = list->next) {
-		AppletData *ad = list->data;
-		if (ad->expand_major && ad->size_hints)
-			ad_with_hints = g_list_prepend (ad_with_hints, ad);
-	}
-
-	/* put the list in the correct order: this is important
-	 * since we'll use this order in the size_allocate() */
-	ad_with_hints = g_list_reverse (ad_with_hints);
-
-	panel->nb_applets_size_hints = g_list_length (ad_with_hints);
-
-	if (panel->nb_applets_size_hints > 0) {
-		int i;
-		panel->applets_hints = g_new0 (AppletSizeHints,
-					       panel->nb_applets_size_hints);
-		panel->applets_using_hint = g_new0 (AppletSizeHintsAlloc,
-						    panel->nb_applets_size_hints);
-
-		i = 0;
-		for (list = ad_with_hints;
-		     list != NULL;
-		     list = g_list_next (list)) {
-			AppletData *ad = list->data;
-
-			panel->applets_hints[i].hints = ad->size_hints;
-			panel->applets_hints[i].len = ad->size_hints_len;
-			i++;
-		}
-	}
-}
-
 /* Note: this can only be called at the beginning of size_allocate, which means
- * that ad->constrained doesn't matter yet (it will be set to the correct
+ * that ad->position doesn't matter yet (it will be set to the correct
  * value in size_allocate). */
 static void
 panel_widget_update_positions_packed_start (PanelWidget *panel)
@@ -619,7 +835,7 @@ panel_widget_update_positions_packed_start (PanelWidget *panel)
 	/* get size used by the objects */
 	for (l = list; l; l = l->next) {
 		ad = l->data;
-		size_all += ad->cells;
+		size_all += ad->size;
 	}
 
 	/* update absolute position of all applets based on this information,
@@ -629,8 +845,8 @@ panel_widget_update_positions_packed_start (PanelWidget *panel)
 
 	while (l) {
 		ad = l->data;
-		ad->constrained = pos_next;
-		pos_next += ad->cells;
+		ad->position = pos_next;
+		pos_next += ad->size;
 		l = l->next;
 	}
 
@@ -639,10 +855,8 @@ panel_widget_update_positions_packed_start (PanelWidget *panel)
 
 /* Note: only use this function when you can; see comment above
  * panel_widget_update_positions_packed_start()
- * For center specifically, we require ad->cells to be set. Note that we don't
- * care much about min_cells: if we need it, this means objects will have to be
- * pushed to accomodate other objects, which will kill centering anyway.
- * (FIXME: hrm, not that sure about it ;-)) */
+ * For center specifically, we require ad->size to be set.
+ */
 static void
 panel_widget_update_positions_packed_center (PanelWidget *panel)
 {
@@ -659,7 +873,7 @@ panel_widget_update_positions_packed_center (PanelWidget *panel)
 	/* get size used by the objects */
 	for (l = list; l; l = l->next) {
 		ad = l->data;
-		size_all += ad->cells;
+		size_all += ad->size;
 	}
 
 	/* update absolute position of all applets based on this information,
@@ -669,8 +883,8 @@ panel_widget_update_positions_packed_center (PanelWidget *panel)
 
 	while (l) {
 		ad = l->data;
-		ad->constrained = pos_next;
-		pos_next += ad->cells;
+		ad->position = pos_next;
+		pos_next += ad->size;
 		l = l->next;
 	}
 
@@ -695,7 +909,7 @@ panel_widget_update_positions_packed_end (PanelWidget *panel)
 	/* get size used by the objects */
 	for (l = list; l; l = l->next) {
 		ad = l->data;
-		size_all += ad->cells;
+		size_all += ad->size;
 	}
 
 	/* update absolute position of all applets based on this information,
@@ -705,8 +919,8 @@ panel_widget_update_positions_packed_end (PanelWidget *panel)
 
 	while (l) {
 		ad = l->data;
-		ad->constrained = pos_next;
-		pos_next += ad->cells;
+		ad->position = pos_next;
+		pos_next += ad->size;
 		l = l->next;
 	}
 
@@ -723,18 +937,13 @@ panel_widget_update_positions (PanelWidget *panel)
 	i = 0;
 
 	if (panel->packed) {
-		/* Because of the following, we have to be very careful that
-		 * this code path is called only before a size_request() on the
-		 * toplevel */
-		panel_widget_update_size_hints_for_toplevel (panel);
-
 		/* keep in sync with code in size_allocate */
 		for (list = panel->applet_list;
 		     list != NULL;
 		     list = g_list_next (list)) {
 			ad = list->data;
-			ad->constrained = i;
-			i += ad->cells;
+			ad->position = i;
+			i += ad->size;
 		}
 	} else {
 		/* Re-compute the ideal position of objects, based on their size */
@@ -748,54 +957,10 @@ panel_widget_update_positions (PanelWidget *panel)
 		     list != NULL;
 		     list = g_list_next (list)) {
 			ad = list->data;
-			if (ad->constrained < i)
-				ad->constrained = i;
+			if (ad->position < i)
+				ad->position = i;
 
-			i = ad->constrained + ad->cells;
-		}
-
-		/* Third pass: now expand from the end, and start using size
-		 * hints if we need more room */
-		i = panel->size;
-		for(list = g_list_last(panel->applet_list);
-		    list!=NULL;
-		    list = g_list_previous(list)) {
-			int cells;
-
-			ad = list->data;
-
-			if (ad->constrained + ad->min_cells > i)
-				ad->constrained = MAX (i - ad->min_cells, 0);
-
-			if (ad->expand_major) {
-				cells = (i - ad->constrained) - 1;
-
-				if (ad->size_hints)
-					cells = get_size_from_hints (ad, cells);
-				cells = MAX (cells, ad->min_cells);
-				cells = MIN (cells, panel->size);
-
-				ad->cells = cells;
-			}
-
-			i = ad->constrained;
-		}
-
-		/* EEEEK, there's not enough room, so shift applets even
-		 * at the expense of perhaps running out of room on the
-		 * right if there is no free space in the middle */
-		if(i < 0) {
-			i = 0;
-			for(list = panel->applet_list;
-			    list!=NULL;
-			    list = g_list_next(list)) {
-				ad = list->data;
-
-				if (ad->constrained < i)
-					ad->constrained = i;
-
-				i = ad->constrained + ad->cells;
-			}
+			i = ad->position + ad->size;
 		}
 	}
 }
@@ -805,7 +970,7 @@ panel_widget_get_moveby (PanelWidget *panel,
 			 AppletData  *ad)
 {
 	/* move relative to the center of the object */
-	return panel_widget_get_cursorloc (panel) - ad->constrained - ad->cells / 2;
+	return panel_widget_get_cursorloc (panel) - ad->position - ad->size / 2;
 }
 
 static int
@@ -838,7 +1003,7 @@ panel_widget_move_get_pos_next_pack (PanelWidget *panel,
 	if (!nad || nad->pack_type > ad->pack_type + 1)
 		return panel_widget_move_get_pos_pack (panel, ad->pack_type + 1);
 
-	return nad->constrained;
+	return nad->position;
 }
 
 static int
@@ -849,7 +1014,7 @@ panel_widget_move_get_pos_prev_pack (PanelWidget *panel,
 	if (!pad || pad->pack_type < ad->pack_type - 1)
 		return panel_widget_move_get_pos_pack (panel, ad->pack_type - 1);
 
-	return pad->constrained + pad->cells;
+	return pad->position + pad->size;
 }
 
 static void
@@ -930,14 +1095,14 @@ panel_widget_switch_applet_right (PanelWidget *panel,
 		return FALSE;
 
 	/* count moveby from end of object => remove distance to go there */
-	moveby -= ad->cells / 2;
+	moveby -= ad->size / 2;
 
 	nad = list->next ? list->next->data : NULL;
 
 	/* Move inside same pack */
 	if (nad && nad->pack_type == ad->pack_type) {
 		if (force_switch ||
-		    (moveby >= nad->cells / 2)) {
+		    (moveby >= nad->size / 2)) {
 			swap_index = ad->pack_index;
 			ad->pack_index = nad->pack_index;
 			nad->pack_index = swap_index;
@@ -958,7 +1123,7 @@ panel_widget_switch_applet_right (PanelWidget *panel,
 	/* Move to next pack */
 	next_pos = panel_widget_move_get_pos_next_pack (panel, ad, nad);
 	if (force_switch ||
-	    (moveby >= (next_pos - (ad->constrained + ad->cells)) / 2)) {
+	    (moveby >= (next_pos - (ad->position + ad->size)) / 2)) {
 		if (ad->pack_type + 1 == PANEL_OBJECT_PACK_END)
 			panel_widget_move_to_pack (panel, ad, ad->pack_type + 1, -1);
 		else
@@ -1025,14 +1190,14 @@ panel_widget_switch_applet_left (PanelWidget *panel,
 		return FALSE;
 
 	/* count moveby from start of object => add distance to go there */
-	moveby += ad->cells / 2;
+	moveby += ad->size / 2;
 
 	pad = list->prev ? list->prev->data : NULL;
 
 	/* Move inside same pack */
 	if (pad && pad->pack_type == ad->pack_type) {
 		if (force_switch ||
-		    (moveby <= - pad->cells / 2)) {
+		    (moveby <= - pad->size / 2)) {
 			swap_index = ad->pack_index;
 			ad->pack_index = pad->pack_index;
 			pad->pack_index = swap_index;
@@ -1053,7 +1218,7 @@ panel_widget_switch_applet_left (PanelWidget *panel,
 	/* Move to prev pack */
 	prev_pos = panel_widget_move_get_pos_prev_pack (panel, ad, pad);
 	if (force_switch ||
-	    (moveby <=  - ((ad->constrained - prev_pos) / 2))) {
+	    (moveby <= - ((ad->position - prev_pos) / 2))) {
 		panel_widget_move_to_pack (panel, ad, ad->pack_type - 1, -1);
 		emit_applet_moved (panel, ad);
 
@@ -1082,16 +1247,16 @@ panel_widget_switch_move (PanelWidget *panel,
 
 	moveby = panel_widget_get_moveby (panel, ad);
 
-	if (moveby > ad->cells / 2) {
+	if (moveby > ad->size / 2) {
 		moved = TRUE;
-		while (moved && moveby > ad->cells / 2) {
+		while (moved && moveby > ad->size / 2) {
 			moved = panel_widget_switch_applet_right (panel, list,
 								  moveby, FALSE);
 			moveby = panel_widget_get_moveby (panel, ad);
 		}
 	} else {
 		moved = TRUE;
-		while (moved && moveby < - ad->cells / 2) {
+		while (moved && moveby < - ad->size / 2) {
 			moved = panel_widget_switch_applet_left (panel, list,
 								 moveby, FALSE);
 			moveby = panel_widget_get_moveby (panel, ad);
@@ -1118,7 +1283,7 @@ panel_widget_push_applet_right (PanelWidget *panel,
 		return FALSE;
 
 	/* count moveby from end of object => remove distance to go there */
-	moveby -= ad->cells / 2;
+	moveby -= ad->size / 2;
 
 	new_pack_type = ad->pack_type + 1;
 
@@ -1134,7 +1299,7 @@ panel_widget_push_applet_right (PanelWidget *panel,
 	next_pos = panel_widget_move_get_pos_next_pack (panel, ad, nad);
 
 	if (!force_switch &&
-	    (moveby < (next_pos - (ad->constrained + ad->cells)) / 2))
+	    (moveby < (next_pos - (ad->position + ad->size)) / 2))
 		return FALSE;
 
 	for (l = last_in_pack; l; l = l->prev) {
@@ -1176,7 +1341,7 @@ panel_widget_push_applet_left (PanelWidget *panel,
 		return FALSE;
 
 	/* count moveby from start of object => add distance to go there */
-	moveby += ad->cells / 2;
+	moveby += ad->size / 2;
 
 	new_pack_type = ad->pack_type - 1;
 
@@ -1192,7 +1357,7 @@ panel_widget_push_applet_left (PanelWidget *panel,
 	prev_pos = panel_widget_move_get_pos_prev_pack (panel, ad, pad);
 
 	if (!force_switch &&
-	    (moveby >  - ((ad->constrained - prev_pos) / 2)))
+	    (moveby > - ((ad->position - prev_pos) / 2)))
 		return FALSE;
 
 	for (l = first_in_pack; l; l = l->next) {
@@ -1230,10 +1395,10 @@ panel_widget_push_move (PanelWidget *panel,
 
 	moveby = panel_widget_get_moveby (panel, ad);
 
-	if (direction > 0 || moveby > ad->cells / 2) {
+	if (direction > 0 || moveby > ad->size / 2) {
 		moved = TRUE;
 		while (direction > 0 ||
-		       (moved && moveby > ad->cells / 2)) {
+		       (moved && moveby > ad->size / 2)) {
 			moved = panel_widget_push_applet_right (panel, list,
 								moveby,
 								direction != 0);
@@ -1246,7 +1411,7 @@ panel_widget_push_move (PanelWidget *panel,
 	} else {
 		moved = TRUE;
 		while (direction < 0 ||
-		       (moved && moveby < ad->cells / 2)) {
+		       (moved && moveby < ad->size / 2)) {
 			moved = panel_widget_push_applet_left (panel, list,
 							       moveby,
 							       direction != 0);
@@ -1259,78 +1424,6 @@ panel_widget_push_move (PanelWidget *panel,
 	}
 }
 
-static void
-panel_widget_size_request(GtkWidget *widget, GtkRequisition *requisition)
-{
-	PanelWidget *panel;
-	GList *list;
-	gboolean dont_fill;
-
-	g_return_if_fail(PANEL_IS_WIDGET(widget));
-	g_return_if_fail(requisition!=NULL);
-
-	panel = PANEL_WIDGET(widget);
-
-	if(panel->orient == GTK_ORIENTATION_HORIZONTAL) {
-		requisition->width = 0;
-		requisition->height = panel->sz;
-	} else {
-		requisition->height = 0;
-		requisition->width = panel->sz;
-	}
-
-	for(list = panel->applet_list; list!=NULL; list = g_list_next(list)) {
-		AppletData *ad = list->data;
-                GtkRequisition child_req;
-
-                gtk_widget_get_preferred_size (ad->applet, &child_req, NULL);
-
-		if (panel->orient == GTK_ORIENTATION_HORIZONTAL) {
-			if (panel->packed &&
-			    !(ad->expand_major && ad->size_hints))
-				requisition->width += child_req.width;
-		} else {
-			if (panel->packed &&
-			    !(ad->expand_major && ad->size_hints))
-				requisition->height += child_req.height;
-		}
-	}
-
-	dont_fill = panel->packed && panel->nb_applets_size_hints != 0;
-
-	if (panel->orient == GTK_ORIENTATION_HORIZONTAL) {
-		if (requisition->width < 12 && !dont_fill)
-			requisition->width = 12;
-		if (requisition->height < 12)
-			requisition->height = 12;
-	} else {
-		if (requisition->width < 12)
-			requisition->width = 12;
-		if (requisition->height < 12 && !dont_fill)
-			requisition->height = 12;
-	}
-}
-
-static void
-panel_widget_get_preferred_width(GtkWidget *widget, gint *minimal_width, gint *natural_width)
-{
-	GtkRequisition requisition;
-
-        panel_widget_size_request (widget, &requisition);
-
-	*minimal_width = *natural_width = requisition.width;
-}
-
-static void
-panel_widget_get_preferred_height(GtkWidget *widget, gint *minimal_height, gint *natural_height)
-{
-	GtkRequisition requisition;
-
-        panel_widget_size_request (widget, &requisition);
-
-	*minimal_height = *natural_height = requisition.height;
-}
-
 static void
 queue_resize_on_all_applets(PanelWidget *panel)
 {
@@ -1343,156 +1436,193 @@ queue_resize_on_all_applets(PanelWidget *panel)
 }
 
 static void
-panel_widget_size_allocate(GtkWidget *widget, GtkAllocation *allocation)
+get_preferred_size (PanelWidget    *self,
+                    AppletData     *ad,
+                    GtkRequisition *minimum_size,
+                    GtkRequisition *natural_size)
 {
-	PanelWidget *panel;
-	GList *list;
-	int i;
-	gboolean ltr;
-	AppletData *ad;
+  if (self->orient == GTK_ORIENTATION_HORIZONTAL)
+    {
+      if (ad->expand_minor)
+        {
+          minimum_size->height = self->sz;
+          natural_size->height = self->sz;
+        }
+      else
+        {
+          gtk_widget_get_preferred_height (ad->applet,
+                                           &minimum_size->height,
+                                           &natural_size->height);
+        }
 
-	g_return_if_fail(PANEL_IS_WIDGET(widget));
-	g_return_if_fail(allocation!=NULL);
+      gtk_widget_get_preferred_width_for_height (ad->applet,
+                                                 minimum_size->height,
+                                                 &minimum_size->width,
+                                                 &natural_size->width);
 
-	panel = PANEL_WIDGET(widget);
+      if (ad->expand_major)
+        {
+          if (ad->size_hints != NULL)
+            {
+              get_preferred_size_from_hints (ad,
+                                             &minimum_size->width,
+                                             &natural_size->width);
+            }
+        }
+      else
+        {
+          natural_size->width = minimum_size->width;
+        }
+    }
+  else if (self->orient == GTK_ORIENTATION_VERTICAL)
+    {
+      if (ad->expand_minor)
+        {
+          minimum_size->width = self->sz;
+          natural_size->width = self->sz;
+        }
+      else
+        {
+          gtk_widget_get_preferred_width (ad->applet,
+                                          &minimum_size->width,
+                                          &natural_size->width);
+        }
 
-	ltr = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_LTR;
-	
-	gtk_widget_set_allocation (widget, allocation);
-	if (gtk_widget_get_realized (widget))
-		gdk_window_move_resize (gtk_widget_get_window (widget),
-					allocation->x, 
-					allocation->y,
-					allocation->width, 
-					allocation->height);
+      gtk_widget_get_preferred_height_for_width (ad->applet,
+                                                 minimum_size->width,
+                                                 &minimum_size->height,
+                                                 &natural_size->height);
 
-	if(panel->orient == GTK_ORIENTATION_HORIZONTAL)
-		panel->size = allocation->width;
-	else
-		panel->size = allocation->height;
+      if (ad->expand_major)
+        {
+          if (ad->size_hints != NULL)
+            {
+              get_preferred_size_from_hints (ad,
+                                             &minimum_size->height,
+                                             &natural_size->height);
+            }
+        }
+      else
+        {
+          natural_size->height = minimum_size->height;
+        }
+    }
+  else
+    {
+      g_assert_not_reached ();
+    }
+}
 
-	if (panel->packed) {
-		/* we're assuming the order is the same as the one that was
-		 * in size_request() */
-		int applet_using_hint_index = 0;
+static void
+panel_widget_size_allocate (GtkWidget     *widget,
+                            GtkAllocation *allocation)
+{
+  PanelWidget *self;
+  int remaining_space;
+  GArray *requested_sizes;
+  GList *l;
+  guint i;
+  gboolean ltr;
 
-		/* keep in sync with code in update_positions */
-		i = 0;
-		for(list = panel->applet_list;
-		    list!=NULL;
-		    list = g_list_next(list)) {
-			GtkAllocation challoc;
-			GtkRequisition chreq;
+  self = PANEL_WIDGET (widget);
 
-			ad = list->data;
+  gtk_widget_set_allocation (widget, allocation);
 
-			gtk_widget_get_preferred_size (ad->applet, &chreq, NULL);
+  if (gtk_widget_get_realized (widget))
+    {
+      gdk_window_move_resize (gtk_widget_get_window (widget),
+                              allocation->x,
+                              allocation->y,
+                              allocation->width,
+                              allocation->height);
+    }
 
-			ad->constrained = i;
-			
-			challoc.width = chreq.width;
-			challoc.height = chreq.height;
-			if(panel->orient == GTK_ORIENTATION_HORIZONTAL) {
-				if (ad->expand_minor)
-					challoc.height = allocation->height;
+  if (self->orient == GTK_ORIENTATION_HORIZONTAL)
+    self->size = allocation->width;
+  else
+    self->size = allocation->height;
 
-				if (ad->expand_major && ad->size_hints) {
-					int width = panel->applets_using_hint[applet_using_hint_index].size;
-					applet_using_hint_index++;
-					challoc.width = MIN (width, allocation->width - i);
-				}
+  remaining_space = self->size;
+  requested_sizes = g_array_new (FALSE, FALSE, sizeof (GtkRequestedSize));
 
-				ad->cells = challoc.width;
-				challoc.x = ltr ? ad->constrained : panel->size - ad->constrained - challoc.width;
-				challoc.y = allocation->height / 2 - challoc.height / 2;
-			} else {
-				if (ad->expand_minor)
-					challoc.width = allocation->width;
+  for (l = self->applet_list; l != NULL; l = g_list_next (l))
+    {
+      AppletData *ad;
+      GtkRequisition minimum;
+      GtkRequisition natural;
+      GtkRequestedSize request;
 
-				if (ad->expand_major && ad->size_hints) {
-					int height = panel->applets_using_hint[applet_using_hint_index].size;
-					applet_using_hint_index++;
-					challoc.height = MIN (height, allocation->height - i);
-				}
+      ad = l->data;
 
-				ad->cells = challoc.height;
-				challoc.x = allocation->width / 2 - challoc.width / 2;
-				challoc.y = ad->constrained;
-			}
-			ad->min_cells  = ad->cells;
+      get_preferred_size (self, ad, &minimum, &natural);
 
-			challoc.width = MAX (challoc.width, 0);
-			challoc.height = MAX (challoc.height, 0);
+      if (self->orient == GTK_ORIENTATION_HORIZONTAL)
+        {
+          request.minimum_size = minimum.width;
+          request.natural_size = natural.width;
+        }
+      else
+        {
+          request.minimum_size = minimum.height;
+          request.natural_size = natural.height;
+        }
 
-			gtk_widget_size_allocate(ad->applet,&challoc);
-			i += ad->cells;
-		}
+      request.data = ad;
+      remaining_space -= request.minimum_size;
 
-		/* EEEEK, there might be not enough room and we don't handle
-		 * it: all the applets at the right well be unusable */
+      g_array_append_val (requested_sizes, request);
+    }
 
-	} else { /*not packed*/
+  gtk_distribute_natural_allocation (remaining_space,
+                                     requested_sizes->len,
+                                     (GtkRequestedSize *) requested_sizes->data);
 
-		/* First pass: set ad->cells so that we can know the absolute
-		 * position of objects. */
-		for (list = panel->applet_list;
-		     list != NULL;
-		     list = g_list_next (list)) {
-			GtkRequisition chreq;
+  for (i = 0; i < requested_sizes->len; i++)
+    {
+      GtkRequestedSize *request;
+      AppletData *ad;
 
-			ad = list->data;
+      request = &g_array_index (requested_sizes, GtkRequestedSize, i);
+      ad = request->data;
 
-			gtk_widget_get_preferred_size (ad->applet, &chreq, NULL);
+      ad->size = request->minimum_size;
+    }
 
-			if (!ad->expand_major || !ad->size_hints) {
-				if(panel->orient == GTK_ORIENTATION_HORIZONTAL)
-					ad->cells = chreq.width;
-				else
-					ad->cells = chreq.height;
+  g_array_free (requested_sizes, TRUE);
+  panel_widget_update_positions (self);
 
-				ad->min_cells = ad->cells;
-			} else {
-				ad->cells = ad->size_hints [0];
-				ad->min_cells = ad->size_hints [ad->size_hints_len - 1];
-			}
-		}
+  ltr = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_LTR;
 
-		panel_widget_update_positions (panel);
+  for (l = self->applet_list; l != NULL; l = g_list_next (l))
+    {
+      AppletData *ad;
+      GtkRequisition minimum;
+      GtkRequisition natural;
+      GtkAllocation challoc;
 
-		/* Last pass: actually allocate the size for each object */
-		for(list = panel->applet_list;
-		    list!=NULL;
-		    list = g_list_next(list)) {
-			GtkAllocation challoc;
-			GtkRequisition chreq;
+      ad = l->data;
 
-			ad = list->data;
+      get_preferred_size (self, ad, &minimum, &natural);
 
-			gtk_widget_get_preferred_size (ad->applet, &chreq, NULL);
-			challoc.width = chreq.width;
-			challoc.height = chreq.height;
-			if(panel->orient == GTK_ORIENTATION_HORIZONTAL) {
-				challoc.width = ad->cells;
-				if (ad->expand_minor) {
-					challoc.height = allocation->height;
-				}
-				challoc.x = ltr ? ad->constrained : panel->size - ad->constrained - challoc.width;
-				challoc.y = allocation->height / 2 - challoc.height / 2;
-			} else {
-				challoc.height = ad->cells;
-				if (ad->expand_minor) {
-					challoc.width = allocation->width;
-				}
-				challoc.x = allocation->width / 2 - challoc.width / 2;
-				challoc.y = ad->constrained;
-			}
+      if (self->orient == GTK_ORIENTATION_HORIZONTAL)
+        {
+          challoc.width = ad->size;
+          challoc.height = minimum.height;
 
-			challoc.width = MAX (challoc.width, 0);
-			challoc.height = MAX (challoc.height, 0);
+          challoc.x = ltr ? ad->position : self->size - ad->position - challoc.width;
+          challoc.y = allocation->height / 2 - challoc.height / 2;
+        }
+      else
+        {
+          challoc.width = minimum.width;
+          challoc.height = ad->size;
 
-			gtk_widget_size_allocate(ad->applet,&challoc);
-		}
-	}
+          challoc.x = allocation->width / 2 - challoc.width / 2;
+          challoc.y = ad->position;
+        }
+
+      gtk_widget_size_allocate (ad->applet, &challoc);
+    }
 }
 
 gboolean
@@ -1530,25 +1660,6 @@ panel_widget_is_cursor(PanelWidget *panel, int overlap)
 	return FALSE;
 }
 
-static void
-panel_widget_finalize (GObject *obj)
-{
-	PanelWidget *panel;
-
-	g_return_if_fail (PANEL_IS_WIDGET (obj));
-
-	panel = PANEL_WIDGET (obj);
-
-	if (panel->applets_hints != NULL)
-		g_free (panel->applets_hints);
-	panel->applets_hints = NULL;
-	if (panel->applets_using_hint != NULL)
-		g_free (panel->applets_using_hint);
-	panel->applets_using_hint = NULL;
-
-	G_OBJECT_CLASS (panel_widget_parent_class)->finalize (obj);
-}
-
 static void
 panel_widget_open_dialog_destroyed (PanelWidget *panel_widget,
 				    GtkWidget   *dialog)
@@ -1591,7 +1702,6 @@ panel_widget_dispose (GObject *obj)
 static void
 panel_widget_init (PanelWidget *panel)
 {
-	GtkStyleContext *context;
 	GtkWidget *widget = (GtkWidget *) panel;
 
 	gtk_widget_set_events (
@@ -1605,13 +1715,6 @@ panel_widget_init (PanelWidget *panel)
 	panel->drop_widget   = widget;
 	panel->open_dialogs  = NULL;
 
-	panel->nb_applets_size_hints = 0;
-	panel->applets_hints = NULL;
-	panel->applets_using_hint = NULL;
-
-	context = gtk_widget_get_style_context (widget);
-	gtk_style_context_add_class (context, GTK_STYLE_CLASS_HORIZONTAL);
-
 	panels = g_slist_append (panels, panel);
 }
 
@@ -1822,8 +1925,8 @@ panel_widget_get_insert_at_cursor (PanelWidget         *widget,
 	for (l = widget->applet_list; l; l = l->next) {
 		ad = l->data;
 
-		if (ad->constrained <= pos) {
-			if (ad->constrained + ad->cells > pos) {
+		if (ad->position <= pos) {
+			if (ad->position + ad->size > pos) {
 				*pack_type = ad->pack_type;
 				*pack_index = ad->pack_index;
 			}
@@ -1889,9 +1992,7 @@ panel_widget_applet_move_to_cursor (PanelWidget *panel)
 				PANEL_WIDGET(list->data);
 
 			if (panel != new_panel &&
-			    panel_widget_is_cursor (new_panel,10) &&
-			    panel_screen_from_panel_widget (panel) ==
-			    panel_screen_from_panel_widget (new_panel)) {
+			    panel_widget_is_cursor (new_panel,10)) {
 				PanelObjectPackType pack_type = PANEL_OBJECT_PACK_START;
 				int                 pack_index = 0;
 
@@ -2307,11 +2408,10 @@ panel_widget_add (PanelWidget         *panel,
 	if (ad == NULL) {
 		ad = g_new (AppletData, 1);
 		ad->applet = applet;
-		ad->cells = 1;
-		ad->min_cells = 1;
+		ad->size = 1;
 		ad->pack_type = pack_type;
 		ad->pack_index = pack_index;
-		ad->constrained = 0;
+		ad->position = 0;
 		ad->expand_major = FALSE;
 		ad->expand_minor = FALSE;
 		ad->size_hints = NULL;
@@ -2411,8 +2511,6 @@ panel_widget_set_packed (PanelWidget *panel_widget,
 {
 	panel_widget->packed = packed;
 
-	panel_widget_update_size_hints_for_toplevel (panel_widget);
-
 	gtk_widget_queue_resize (GTK_WIDGET (panel_widget));
 }
 
@@ -2420,19 +2518,8 @@ void
 panel_widget_set_orientation (PanelWidget    *panel_widget,
 			      GtkOrientation  orientation)
 {
-	GtkStyleContext *context;
-
 	panel_widget->orient = orientation;
 
-	context = gtk_widget_get_style_context (GTK_WIDGET (panel_widget));
-	if (orientation == GTK_ORIENTATION_HORIZONTAL) {
-		gtk_style_context_add_class (context, GTK_STYLE_CLASS_HORIZONTAL);
-		gtk_style_context_remove_class (context, GTK_STYLE_CLASS_VERTICAL);
-	} else {
-		gtk_style_context_add_class (context, GTK_STYLE_CLASS_VERTICAL);
-		gtk_style_context_remove_class (context, GTK_STYLE_CLASS_HORIZONTAL);
-	}
-
 	gtk_widget_queue_resize (GTK_WIDGET (panel_widget));
 }
 
@@ -2645,7 +2732,6 @@ panel_widget_set_applet_size_hints (PanelWidget *panel,
 		ad->size_hints = NULL;
 	}
 
-	panel_widget_update_size_hints_for_toplevel (panel);
 	gtk_widget_queue_resize (GTK_WIDGET (panel));
 }
 
