@@ -33,19 +33,6 @@ typedef struct _PanelWidgetClass	PanelWidgetClass;
 
 typedef struct _AppletData		AppletData;
 
-typedef struct _AppletSizeHints		AppletSizeHints;
-typedef struct _AppletSizeHintsAlloc	AppletSizeHintsAlloc;
-
-struct _AppletSizeHints {
-	int *hints;
-	int  len;
-};
-
-struct _AppletSizeHintsAlloc {
-	int index;
-	int size;
-};
-
 struct _AppletData
 {
 	GtkWidget *	applet;
@@ -53,11 +40,8 @@ struct _AppletData
 	PanelObjectPackType pack_type;
 	int		    pack_index;
 
-	/* constrained position from the left, after taking other objects into
-	 * account */
-	int             constrained;
-	int		cells;
-	int             min_cells;
+	int             position;
+	int             size;
 
 	/* Valid size ranges for expanded applets */
 	int *           size_hints; 
@@ -89,12 +73,6 @@ struct _PanelWidget
 	
 	PanelToplevel  *toplevel;
 
-	/* helpers to get a good size in packed panels with applets using
-	 * size hints */
-	int                   nb_applets_size_hints;
-	AppletSizeHints      *applets_hints;
-	AppletSizeHintsAlloc *applets_using_hint;
-
 	guint           packed : 1;
 };
 
