@@ -339,13 +339,12 @@ something_fishy_going_on (FishApplet *fish,
 
 static gboolean
 locate_fortune_command (FishApplet   *fish,
-			int          *argcp,
-			char       ***argvp)
+                        char       ***argvp)
 {
 	char *prog = NULL;
 
 	if (fish->command
-	    && g_shell_parse_argv (fish->command, argcp, argvp, NULL)) {
+	    && g_shell_parse_argv (fish->command, NULL, argvp, NULL)) {
 		prog = g_find_program_in_path ((*argvp)[0]);
 		if (prog) {
 			g_free (prog);
@@ -353,22 +352,23 @@ locate_fortune_command (FishApplet   *fish,
 		}
 
 		g_strfreev (*argvp);
+		*argvp = NULL;
 	}
 
 	prog = g_find_program_in_path ("fortune");
 	if (prog) {
 		g_free (prog);
-		if (g_shell_parse_argv ("fortune", argcp, argvp, NULL))
+		if (g_shell_parse_argv ("fortune", NULL, argvp, NULL))
 			return FALSE;
 	}
 
 	if (g_file_test ("/usr/games/fortune", G_FILE_TEST_IS_EXECUTABLE)
-	    && g_shell_parse_argv ("/usr/games/fortune", argcp, argvp, NULL))
+	    && g_shell_parse_argv ("/usr/games/fortune", NULL, argvp, NULL))
 		return FALSE;
 
 	something_fishy_going_on (fish,
 				  _("Unable to locate the command to execute"));
-	*argvp = NULL;
+
 	return FALSE;
 }
 
@@ -541,7 +541,6 @@ display_fortune_dialog (FishApplet *fish)
 	gboolean     user_command;
 	int          output;
 	const char  *charset;
-	int          argc;
 	char       **argv;
 
 	/* if there is still a pipe, close it */
@@ -550,8 +549,10 @@ display_fortune_dialog (FishApplet *fish)
 	fish->source_id = 0;
 	fish_close_channel (fish);
 
-	user_command = locate_fortune_command (fish, &argc, &argv);
-	if (!argv)
+	argv = NULL;
+	user_command = locate_fortune_command (fish, &argv);
+
+	if (argv == NULL)
 		return;
 
 	if (!fish->fortune_dialog) {
