@@ -29,11 +29,15 @@ struct _GpMenu
   GtkMenu                parent;
 
   gchar                 *name;
-  GpApplet              *applet;
   gboolean               required;
 
+  gboolean               enable_tooltips;
+  gboolean               locked_down;
+  guint                  menu_icon_size;
+
   GMenuTree             *tree;
 
+  gboolean               loaded;
   gboolean               empty;
 
   GpAppendMenuItemsFunc  append_func;
@@ -41,18 +45,29 @@ struct _GpMenu
 
   guint                  reload_id;
 
-  gulong                 locked_down_id;
-  gulong                 menu_icon_size_id;
+  gchar                 *path;
 };
 
+enum
+{
+  LOADED,
+
+  LAST_SIGNAL
+};
+
+static guint menu_signals[LAST_SIGNAL] = { 0 };
+
 enum
 {
   PROP_0,
 
   PROP_NAME,
-  PROP_APPLET,
   PROP_REQUIRED,
 
+  PROP_ENABLE_TOOLTIPS,
+  PROP_LOCKED_DOWN,
+  PROP_MENU_ICON_SIZE,
+
   PROP_EMPTY,
 
   LAST_PROP
@@ -107,13 +122,11 @@ append_directory (GtkMenuShell  *shell,
   if (icon != NULL)
     {
       GtkWidget *image;
-      guint icon_size;
 
       image = gtk_image_new ();
-      icon_size = gp_applet_get_menu_icon_size (menu->applet);
 
       gtk_image_set_from_gicon (GTK_IMAGE (image), icon, GTK_ICON_SIZE_MENU);
-      gtk_image_set_pixel_size (GTK_IMAGE (image), icon_size);
+      gtk_image_set_pixel_size (GTK_IMAGE (image), menu->menu_icon_size);
 
       gp_image_menu_item_set_image (GP_IMAGE_MENU_ITEM (item), image);
     }
@@ -170,13 +183,11 @@ append_entry (GtkMenuShell  *shell,
   if (icon != NULL)
     {
       GtkWidget *image;
-      guint icon_size;
 
       image = gtk_image_new ();
-      icon_size = gp_applet_get_menu_icon_size (menu->applet);
 
       gtk_image_set_from_gicon (GTK_IMAGE (image), icon, GTK_ICON_SIZE_MENU);
-      gtk_image_set_pixel_size (GTK_IMAGE (image), icon_size);
+      gtk_image_set_pixel_size (GTK_IMAGE (image), menu->menu_icon_size);
 
       gp_image_menu_item_set_image (GP_IMAGE_MENU_ITEM (item), image);
     }
@@ -188,13 +199,13 @@ append_entry (GtkMenuShell  *shell,
     {
       gtk_widget_set_tooltip_text (item, description);
 
-      g_object_bind_property (menu->applet, "enable-tooltips",
+      g_object_bind_property (menu, "enable-tooltips",
                               item, "has-tooltip",
                               G_BINDING_DEFAULT |
                               G_BINDING_SYNC_CREATE);
     }
 
-  if (!gp_applet_get_locked_down (menu->applet))
+  if (!menu->locked_down)
     {
       static const GtkTargetEntry drag_targets[] =
         {
@@ -290,14 +301,13 @@ static void
 menu_reload (GpMenu *menu)
 {
   GError *error;
-  gboolean loaded;
   GList *children;
   gboolean empty;
 
   gtk_container_foreach (GTK_CONTAINER (menu), remove_item, NULL);
 
   error = NULL;
-  loaded = gmenu_tree_load_sync (menu->tree, &error);
+  menu->loaded = gmenu_tree_load_sync (menu->tree, &error);
 
   if (error != NULL)
     {
@@ -307,16 +317,27 @@ menu_reload (GpMenu *menu)
       g_clear_error (&error);
     }
 
-  if (loaded)
+  if (menu->loaded)
     {
+      const gchar *path;
       GMenuTreeDirectory *directory;
 
-      directory = gmenu_tree_get_directory_from_path (menu->tree, "/");
+      path = menu->path && *menu->path != '\0' ? menu->path : "/";
+      directory = gmenu_tree_get_directory_from_path (menu->tree, path);
+
+      if (directory == NULL)
+        {
+          g_warning ("Menu path '%s' does not exist!", path);
+          directory = gmenu_tree_get_directory_from_path (menu->tree, "/");
+        }
+
       directory_to_menu_items (directory, GTK_WIDGET (menu), menu);
       gmenu_tree_item_unref (directory);
 
       if (menu->append_func != NULL)
         menu->append_func (GTK_MENU (menu), menu->append_data);
+
+      g_signal_emit (menu, menu_signals[LOADED], 0);
     }
 
   children = gtk_container_get_children (GTK_CONTAINER (menu));
@@ -364,22 +385,6 @@ menu_tree_changed_cb (GMenuTree *tree,
   queue_reload (menu);
 }
 
-static void
-locked_down_cb (GpApplet   *applet,
-                GParamSpec *pspec,
-                GpMenu     *menu)
-{
-  queue_reload (menu);
-}
-
-static void
-menu_icon_size_cb (GpApplet   *applet,
-                   GParamSpec *pspec,
-                   GpMenu     *menu)
-{
-  queue_reload (menu);
-}
-
 static void
 gp_menu_constructed (GObject *object)
 {
@@ -396,14 +401,6 @@ gp_menu_constructed (GObject *object)
   g_signal_connect (menu->tree, "changed",
                     G_CALLBACK (menu_tree_changed_cb), menu);
 
-  menu->locked_down_id = g_signal_connect (menu->applet, "notify::locked-down",
-                                           G_CALLBACK (locked_down_cb), menu);
-
-  menu->menu_icon_size_id = g_signal_connect (menu->applet,
-                                              "notify::menu-icon-size",
-                                              G_CALLBACK (menu_icon_size_cb),
-                                              menu);
-
   queue_reload (menu);
 }
 
@@ -422,20 +419,6 @@ gp_menu_dispose (GObject *object)
       menu->reload_id = 0;
     }
 
-  if (menu->locked_down_id != 0)
-    {
-      g_signal_handler_disconnect (menu->applet, menu->locked_down_id);
-      menu->locked_down_id = 0;
-    }
-
-  if (menu->menu_icon_size_id != 0)
-    {
-      g_signal_handler_disconnect (menu->applet, menu->menu_icon_size_id);
-      menu->menu_icon_size_id = 0;
-    }
-
-  menu->applet = NULL;
-
   G_OBJECT_CLASS (gp_menu_parent_class)->dispose (object);
 }
 
@@ -464,17 +447,16 @@ gp_menu_get_property (GObject    *object,
   switch (property_id)
     {
       case PROP_NAME:
-        g_assert_not_reached ();
-        break;
-
-      case PROP_APPLET:
-        g_assert_not_reached ();
-        break;
-
       case PROP_REQUIRED:
+      case PROP_LOCKED_DOWN:
+      case PROP_MENU_ICON_SIZE:
         g_assert_not_reached ();
         break;
 
+      case PROP_ENABLE_TOOLTIPS:
+        g_value_set_boolean (value, menu->enable_tooltips);
+        break;
+
       case PROP_EMPTY:
         g_value_set_boolean (value, menu->empty);
         break;
@@ -485,6 +467,41 @@ gp_menu_get_property (GObject    *object,
     }
 }
 
+static void
+set_enable_tooltips (GpMenu   *menu,
+                     gboolean  enable_tooltips)
+{
+  if (menu->enable_tooltips == enable_tooltips)
+    return;
+
+  menu->enable_tooltips = enable_tooltips;
+
+  g_object_notify_by_pspec (G_OBJECT (menu),
+                            menu_properties[PROP_ENABLE_TOOLTIPS]);
+}
+
+static void
+set_locked_down (GpMenu    *menu,
+                 gboolean  locked_down)
+{
+  if (menu->locked_down == locked_down)
+    return;
+
+  menu->locked_down = locked_down;
+  queue_reload (menu);
+}
+
+static void
+set_menu_icon_size (GpMenu *menu,
+                    guint   menu_icon_size)
+{
+  if (menu->menu_icon_size == menu_icon_size)
+    return;
+
+  menu->menu_icon_size = menu_icon_size;
+  queue_reload (menu);
+}
+
 static void
 gp_menu_set_property (GObject      *object,
                       guint         property_id,
@@ -502,15 +519,22 @@ gp_menu_set_property (GObject      *object,
         menu->name = g_value_dup_string (value);
         break;
 
-      case PROP_APPLET:
-        g_assert (menu->applet == NULL);
-        menu->applet = g_value_get_object (value);
-        break;
-
       case PROP_REQUIRED:
         menu->required = g_value_get_boolean (value);
         break;
 
+      case PROP_ENABLE_TOOLTIPS:
+        set_enable_tooltips (menu, g_value_get_boolean (value));
+        break;
+
+      case PROP_LOCKED_DOWN:
+        set_locked_down (menu, g_value_get_boolean (value));
+        break;
+
+      case PROP_MENU_ICON_SIZE:
+        set_menu_icon_size (menu, g_value_get_uint (value));
+        break;
+
       case PROP_EMPTY:
         g_assert_not_reached ();
         break;
@@ -530,18 +554,33 @@ install_properties (GObjectClass *object_class)
                          G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE |
                          G_PARAM_STATIC_STRINGS);
 
-  menu_properties[PROP_APPLET] =
-    g_param_spec_object ("applet", "Applet", "Applet",
-                         GP_TYPE_APPLET,
-                         G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE |
-                         G_PARAM_STATIC_STRINGS);
-
   menu_properties[PROP_REQUIRED] =
     g_param_spec_boolean ("required", "Required", "Required",
                           TRUE,
                           G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE |
                           G_PARAM_STATIC_STRINGS);
 
+  menu_properties[PROP_ENABLE_TOOLTIPS] =
+    g_param_spec_boolean ("enable-tooltips", "Enable Tooltips", "Enable Tooltips",
+                          TRUE,
+                          G_PARAM_CONSTRUCT | G_PARAM_READWRITE |
+                          G_PARAM_EXPLICIT_NOTIFY |
+                          G_PARAM_STATIC_STRINGS);
+
+  menu_properties[PROP_LOCKED_DOWN] =
+    g_param_spec_boolean ("locked-down", "Locked Down", "Locked Down",
+                          FALSE,
+                          G_PARAM_CONSTRUCT | G_PARAM_WRITABLE |
+                          G_PARAM_EXPLICIT_NOTIFY |
+                          G_PARAM_STATIC_STRINGS);
+
+  menu_properties[PROP_MENU_ICON_SIZE] =
+    g_param_spec_uint ("menu-icon-size", "Menu Icon Size", "Menu Icon Size",
+                       16, 24, 16,
+                       G_PARAM_CONSTRUCT | G_PARAM_WRITABLE |
+                       G_PARAM_EXPLICIT_NOTIFY |
+                       G_PARAM_STATIC_STRINGS);
+
   menu_properties[PROP_EMPTY] =
     g_param_spec_boolean ("empty", "Empty", "Empty",
                           TRUE,
@@ -551,6 +590,14 @@ install_properties (GObjectClass *object_class)
   g_object_class_install_properties (object_class, LAST_PROP, menu_properties);
 }
 
+static void
+install_signals (void)
+{
+  menu_signals[LOADED] =
+    g_signal_new ("loaded", GP_TYPE_MENU, G_SIGNAL_RUN_LAST,
+                  0, NULL, NULL, NULL, G_TYPE_NONE, 0);
+}
+
 static void
 gp_menu_class_init (GpMenuClass *menu_class)
 {
@@ -565,6 +612,7 @@ gp_menu_class_init (GpMenuClass *menu_class)
   object_class->set_property = gp_menu_set_property;
 
   install_properties (object_class);
+  install_signals ();
 }
 
 static void
@@ -573,17 +621,55 @@ gp_menu_init (GpMenu *menu)
 }
 
 GtkWidget *
-gp_menu_new (GpApplet    *applet,
-             const gchar *name,
+gp_menu_new (const gchar *name,
              gboolean     required)
 {
   return g_object_new (GP_TYPE_MENU,
-                       "applet", applet,
                        "name", name,
                        "required", required,
                        NULL);
 }
 
+void
+gp_menu_set_path (GpMenu      *menu,
+                  const gchar *path)
+{
+  g_free (menu->path);
+  menu->path = g_strdup (path);
+
+  queue_reload (menu);
+}
+
+GIcon *
+gp_menu_get_icon (GpMenu *menu)
+{
+  const gchar *path;
+  GMenuTreeDirectory *directory;
+  GIcon *icon;
+
+  if (!menu->loaded)
+    return NULL;
+
+  path = menu->path && *menu->path != '\0' ? menu->path : "/";
+  directory = gmenu_tree_get_directory_from_path (menu->tree, path);
+
+  if (directory == NULL)
+    return NULL;
+
+  icon = gmenu_tree_directory_get_icon (directory);
+
+  if (icon == NULL)
+    {
+      gmenu_tree_item_unref (directory);
+      return NULL;
+    }
+
+  g_object_ref (icon);
+  gmenu_tree_item_unref (directory);
+
+  return icon;
+}
+
 void
 gp_menu_set_append_func (GpMenu                *menu,
                          GpAppendMenuItemsFunc  append_func,
@@ -594,3 +680,9 @@ gp_menu_set_append_func (GpMenu                *menu,
 
   queue_reload (menu);
 }
+
+void
+gp_menu_reload (GpMenu *menu)
+{
+  menu_reload (menu);
+}
